module MultiSpec where

import qualified Maybe

type Op = String
type Var = String
type BVar = String
data BT
    = S | D | B BVar | L BT BT
      deriving Show

data Exp
    = EVar Var
    | ELam BT Var Exp
    | EApp BT Exp Exp
    | EInt Int 
    | EBool Bool
    | EOp BT Op [Exp]
    | EIf BT Exp Exp Exp
    | ELet BT Var Exp Exp
    | EFix BT Var Exp

    | BLam BVar Exp
    | BApp Exp BT
    | ELift BT BT Exp
      deriving Show


data PEVal
    = PEInt Int
    | PEBool Bool
    | PEFun (PEVal -> PEVal)
    | PEExp Exp
    | PEBT BT

unExp :: PEVal -> Exp
unExp (PEExp e) = e

unFun :: PEVal -> (PEVal -> PEVal)
unFun (PEFun f) = f

unBool :: PEVal -> Bool
unBool (PEBool b) = b

unBV :: PEVal -> BT
unBV (PEBT bt) = bt

instance Show PEVal where
  showsPrec n (PEInt i) = showsPrec n i
  showsPrec n (PEBool i) = showsPrec n i
  showsPrec n (PEFun _) = showString "<function>"
  showsPrec n (PEExp e) = showsPrec n e
  showsPrec n (PEBT b) = showsPrec n b

type OEnv = [(Op, [PEVal] -> PEVal)]
type VEnv = [(Var, PEVal)]
type BEnv = [(BVar, BT)]

lub :: BT -> BT -> BT
lub S bt = bt
lub D bt = D
lub bt S = bt
lub bt D = D
lub bt1 bt2 = L bt1 bt2

btval :: BEnv -> BT -> BT -- always returns S or D
btval b S = S
btval b D = D
btval b (B bv) = Maybe.fromJust $ lookup bv b
btval b (L bt1 bt2) = lub (btval b bt1) (btval b bt2)

spec :: OEnv -> VEnv -> BEnv -> Exp -> PEVal
spec o v b e =
    case e of
      EVar x ->
          Maybe.fromJust $ lookup x v
      ELam bt x e ->
          case btval b bt of
            S -> PEFun (\y -> spec o ((x,y):v) b e)
            D -> PEExp (ELam S x $ unExp $ spec o ((x, PEExp $ EVar x):v) b e)
      EApp bt f e ->
          case btval b bt of
            S -> unFun (spec o v b f) (spec o v b e)
            D -> PEExp (EApp S (unExp $ spec o v b f) (unExp $ spec o v b e))
      EInt i ->
          PEInt i
      EBool b ->
          PEBool b
      EOp bt op xs ->
          case btval b bt of
            S -> Maybe.fromJust (lookup op o) $ map (spec o v b) xs
            D -> PEExp (EOp S op (map (unExp . spec o v b) xs))
      EIf bt c t e ->
          case btval b bt of
            S -> if (unBool $ spec o v b c) then spec o v b t else spec o v b e
            D -> PEExp (EIf S (unExp $ spec o v b c) (unExp $ spec o v b t) (unExp $ spec o v b e))
      ELet bt x e f ->
          case btval b bt of
	    S -> let y = spec o v b e in spec o ((x,y):v) b f
	    D -> PEExp (ELet S x (unExp $ spec o v b e) 
				 (unExp $ spec o ((x,PEExp $ EVar x):v) b f))
      EFix bt x e ->
	  case btval b bt of
	    S -> fix (\y -> spec o ((x,y):v) b e) 
	      where fix f = f (fix f)
	    D -> PEExp (EFix S x (unExp $ spec o ((x, PEExp $ EVar x):v) b e))
      BLam bv e ->
	  PEFun (\y -> spec o v ((bv,unBV y):b) e)
      BApp e bt ->
	  unFun (spec o v b e) (PEBT (btval b bt))
      ELift bt1 bt2 e ->
	  case (btval b bt1, btval b bt2) of
            (S, S) -> spec o v b e
	    (D, S) -> (case spec o v b e of
                         PEInt i -> PEExp (EInt i)
			 PEBool i -> PEExp (EBool i))
	    (D, D) -> spec o v b e

opEq :: [PEVal] -> PEVal
opEq [PEInt i, PEInt j] = PEBool (i==j)

opAdd :: [PEVal] -> PEVal
opAdd [PEInt i, PEInt j] = PEInt (i+j)

opSub :: [PEVal] -> PEVal
opSub [PEInt i, PEInt j] = PEInt (i-j)

initial_o = [("==", opEq), ("+", opAdd), ("-", opSub)]

-- example from Dussart, Henglein, Mossin
exDHM =
    EFix S "f" $
    BLam "bx" $ BLam "by" $ BLam "b" $
    ELam S "x" $ ELam S "y" $
    EIf (B "bx")
        (EOp (B "bx") "==" [EVar"x", ELift (B "bx") S $ EInt 0])
        (ELift (B "b") S $ EInt 1)
        (EApp S (EApp S (BApp (BApp (BApp (EVar "f") (B "by")) (B "bx")) (B "b")) (EVar "y")) (EOp (B "bx") "-" [EVar "x", ELift (B "bx") S $ EInt 1]))

exDHMapplied =
    EApp S (EApp S (BApp (BApp (BApp exDHM S) D) D) (EInt 3)) (EVar "d")

exDHMinitial_v = [("d", PEExp (EVar "d"))]

exDHMrun = spec initial_o exDHMinitial_v [] exDHMapplied


