{-# LANGUAGE TypeFamilies, GADTs, MultiParamTypeClasses, FlexibleInstances, FlexibleContexts, Rank2Types #-}

module PolyBase where

-- :set -XTypeFamilies -XGADTs -XMultiParamTypeClasses -XFlexibleInstances
-- -XFlexibleContexts

import qualified Gen

data S = S
data D = D

data Op a = Op { opname :: String, opvalue :: a }

data Exp t where
    EVar :: String -> Exp t
    EBool :: Bool -> Exp Bool
    EInt :: Int -> Exp Int
    EIf :: Exp Bool -> Exp t -> Exp t -> Exp t
    EConst :: Op t -> Exp t
    EOp1 :: Op (a -> t) -> Exp a -> Exp t
    EOp2 :: Op (a -> b -> t) -> Exp a -> Exp b -> Exp t
    ELam :: (Exp a -> Exp b) -> Exp (a -> b)
    EApp :: Exp (a -> b) -> Exp a -> Exp b
    EFix :: Exp (a -> a) -> Exp a
    ENil :: Exp [a]
    ECons :: Exp a -> Exp [a] -> Exp [a]
    EHead :: Exp [a] -> Exp a
    ETail :: Exp [a] -> Exp [a]
    ENull :: Exp [a] -> Exp Bool

toShowS :: Exp t -> ShowS
toShowS (EBool x) = showString "EBool " . shows x
toShowS (EInt  x) = showString "EInt " . shows x
toShowS (EIf x y z) = showString "EIf (" . toShowS x . 
	                    showString ") (" . toShowS y .
			    showString ") (" . toShowS z .
			    showString ")"
toShowS (EConst op) = showString "EConst " . showString (opname op)
toShowS (EOp1 op x) = showString "EOp1 " . showString (opname op) .
	               showString " (" . toShowS x . showString ")"
toShowS (EOp2 op x y) = showString "EOp2 " . showString (opname op) .
	                 showString " (" . toShowS x . 
			 showString ") (" . toShowS y . showString ")"
toShowS (EVar x) = showString "EVar " . showString x
toShowS (ELam f) = let xx = Gen.string "x" in
		   showString "ELam (\\" . showString xx . showString " -> " .
		     toShowS (f (EVar xx)) .
                    showString ")"
toShowS (EApp f x) = showString "EApp (" . toShowS f . showString ") (" .
                      toShowS x . showString ")"
toShowS (EFix f) = showString "EFix (" . toShowS f . showString ")"
toShowS (ENil) = showString "ENil"
toShowS (ECons hd tl) = showString "ECons (" . toShowS hd . showString ") (" . toShowS tl . showString ")"
toShowS (EHead l) = showString "EHead (" . toShowS l . showString ")"
toShowS (ETail l) = showString "ETail (" . toShowS l . showString ")"
toShowS (ENull l) = showString "ENull (" . toShowS l . showString ")"

simplify :: Exp t -> Exp t
simplify (EApp (EApp (EConst op) x) y) = EOp2 op (simplify x) (simplify y)
simplify (EApp (EConst op) x) = EOp1 op (simplify x)
-- simplify (EOp2 (Op "(+)" _ :: Op (Int -> Int -> Int)) x (EInt 0)) = x
-- simplify (EOp2 op (EInt x) (EInt y)) = EInt (opvalue op x y)
simplify (EIf (EBool True) x y) = simplify x
simplify (EIf (EBool False) x y) = simplify y
simplify x@(EVar _) = x
simplify x@(EBool _) = x
simplify x@(EInt _) = x
simplify (EIf x y z) = (EIf (simplify x) (simplify y) (simplify z))
simplify x@(EConst _) = x
simplify (EOp1 o x) = (EOp1 o (simplify x))
simplify (EOp2 o x y) = (EOp2 o (simplify x) (simplify y))
simplify (ELam f) = (ELam $ \z -> simplify (f z))
simplify (EApp x y) = (EApp (simplify x) (simplify y))
simplify (EFix f) = (EFix (simplify f))
simplify x@(ENil) = x
simplify (ECons x y) = (ECons (simplify x) (simplify y))
simplify (EHead x) = (EHead (simplify x))
simplify (ETail x) = (ETail (simplify x))
simplify (ENull x) = (ENull (simplify x))


simplified :: Exp t -> String
simplified e = toShowS (simplify e) ""

toString :: Exp t -> String
toString e = toShowS e ""

oEq :: Eq a => Op (a -> a -> Bool)
oEq = Op "(==)" (==)
-- 
oEqInt :: Op (Int -> Int -> Bool)
oEqInt = Op "(==)" (==)
--oMult :: Num a => Op (a -> a -> a)
oMult :: Op (Int -> Int -> Int)
oMult = Op "(*)" (*)
--oMinus :: Num a => Op (a -> a -> a)
oMinus :: Op (Int -> Int -> Int)
oMinus = Op "(-)" (-)
--oPlus :: Num a => Op (a -> a -> a)
oPlus :: Op (Int -> Int -> Int)
oPlus = Op "(+)" (+)

oInc :: Op (Int -> Int)
oInc = Op "(+1)" (+1)
